import pandas as pd


def id_slug(a: int, b: int) -> str:
    return f"{int(min(a, b))}-{int(max(a, b))}"  # matches duckdb implementation in 000__schema.sql


def check_required_columns(df: pd.DataFrame, required_columns: list[str]) -> None:
    missing_columns = set(required_columns) - set(df.columns)
    if len(missing_columns) > 0:
        raise ValueError(f"Missing {len(missing_columns)} required column(s): {missing_columns}")



from contextlib import contextmanager
from pathlib import Path

import duckdb

DATABASE_FILE = Path.cwd() / "autoarena.duckdb"
MIGRATION_DIRECTORY = Path(__file__).parent / "migration"

_DATA_DIRECTORY = Path.cwd() / "data"


def get_data_directory() -> Path:
    return _DATA_DIRECTORY


# TODO: allow command line updates? e.g. --data-dir ./my-data
def set_data_directory(path: Path) -> None:
    global _DATA_DIRECTORY
    _DATA_DIRECTORY = path


@contextmanager
def get_database_connection(path: Path) -> duckdb.DuckDBPyConnection:
    conn = duckdb.connect(str(path))
    try:
        yield conn
    finally:
        conn.close()


def get_available_migrations() -> list[Path]:
    return sorted(MIGRATION_DIRECTORY.glob("[0-9][0-9][0-9]__*.sql"))



import os
from abc import abstractmethod, ABCMeta
from typing import Optional

import numpy as np
from loguru import logger


class AutomatedJudge(metaclass=ABCMeta):
    API_KEY_NAME: Optional[str] = None  # if set, verify that this exists in environment on init
    MAX_TOKENS: int = 12  # should really just need one or two
    SLOW_THRESHOLD_SECONDS: float = 5

    _name: str
    _model_name: str
    _system_prompt: str

    n_requests: int
    total_input_tokens: int
    total_output_tokens: int
    response_seconds: list[float]

    def __init__(self, name: str, model_name: str, system_prompt: str):
        self._name = name
        self._model_name = model_name
        self._system_prompt = system_prompt
        self.n_requests = 0
        self.total_input_tokens = 0
        self.total_output_tokens = 0
        self.response_seconds = []
        key = os.environ.get(self.API_KEY_NAME, None) if self.API_KEY_NAME is not None else None
        if self.API_KEY_NAME is not None and key is None:
            message = f"API key '{self.API_KEY_NAME}' must be set in environment running AutoArena to use '{self.name}'"
            raise RuntimeError(message)

    @property
    def name(self) -> str:
        return self._name

    @property
    def model_name(self) -> str:
        return self._model_name

    @property
    def system_prompt(self) -> str:
        return self._system_prompt

    @abstractmethod
    def judge(self, prompt: str, response_a: str, response_b: str) -> str:
        raise NotImplementedError

    @staticmethod
    def verify_environment() -> None:
        """
        Verify that the current environment contains any necessary configuration, such as API keys, necessary to
        interact with this judge. Throw an exception if not.
        """

    def update_usage(self, input_tokens: int, output_tokens: int, response_seconds: float) -> None:
        """Optionally call in `AutomatedJudge.judge` implementations to record usage metrics."""
        self.n_requests += 1
        self.total_input_tokens += input_tokens
        self.total_output_tokens += output_tokens
        self.response_seconds.append(response_seconds)
        if response_seconds >= self.SLOW_THRESHOLD_SECONDS:
            logger.warning(f"Slow response from '{self.name}': {response_seconds:0.3f} seconds")

    def get_usage_summary(self) -> list[str]:
        if self.n_requests == 0 or len(self.response_seconds) == 0:
            return [f"'{self.name}' has not recorded any usage"]
        return [
            f"'{self.name}' used {self.total_input_tokens} input tokens and {self.total_output_tokens} output tokens "
            f"over {self.n_requests} requests",
            f"  * p50 latency: {np.percentile(self.response_seconds, 50):0.3f} seconds",
            f"  * p90 latency: {np.percentile(self.response_seconds, 90):0.3f} seconds",
            f"  * p99 latency: {np.percentile(self.response_seconds, 99):0.3f} seconds",
        ]



import concurrent
import random
from abc import ABCMeta, abstractmethod
from concurrent.futures import ThreadPoolExecutor
from types import TracebackType
from typing import Iterator, TypeVar, Optional

from autoarena.api import api
from autoarena.judge.base import AutomatedJudge

T = TypeVar("T", bound="JudgeExecutor")


class JudgeExecutor(metaclass=ABCMeta):
    def __enter__(self: T) -> T:
        return self

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None: ...

    @abstractmethod
    def execute(
        self, judges_with_head_to_heads: list[tuple[AutomatedJudge, list[api.HeadToHead]]]
    ) -> Iterator[tuple[AutomatedJudge, api.HeadToHead, str]]:
        """Yield responses (winners) from judges as they are ready"""


class BlockingExecutor(JudgeExecutor):
    def execute(
        self, judges_with_head_to_heads: list[tuple[AutomatedJudge, list[api.HeadToHead]]]
    ) -> Iterator[tuple[AutomatedJudge, api.HeadToHead, str]]:
        for judge, head_to_heads in judges_with_head_to_heads:
            for h2h in head_to_heads:
                yield judge, h2h, judge.judge(h2h.prompt, h2h.response_a, h2h.response_b)


class ThreadedExecutor(JudgeExecutor):
    def __init__(self, max_workers: int) -> None:
        self.pool = ThreadPoolExecutor(max_workers=max_workers)

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        self.pool.shutdown(wait=False, cancel_futures=True)

    def execute(
        self, judges_with_head_to_heads: list[tuple[AutomatedJudge, list[api.HeadToHead]]]
    ) -> Iterator[tuple[AutomatedJudge, api.HeadToHead, str]]:
        h2h_with_judges = [(h2h, judge) for judge, h2hs in judges_with_head_to_heads for h2h in h2hs]
        random.shuffle(h2h_with_judges)  # ensure that we are running all judges concurrently to best load balance

        def run(h2h_with_judge: tuple[api.HeadToHead, AutomatedJudge]) -> tuple[AutomatedJudge, api.HeadToHead, str]:
            h, j = h2h_with_judge
            return j, h, j.judge(h.prompt, h.response_a, h.response_b)

        futures = [self.pool.submit(run, h2h_w_j) for h2h_w_j in h2h_with_judges]
        for future in concurrent.futures.as_completed(futures):
            judge, h2h, winner = future.result()
            yield judge, h2h, winner



import functools
import math
import sys
import time
from typing import Callable

if sys.version_info[:2] >= (3, 10):
    from typing import ParamSpec, TypeVar
else:
    from typing_extensions import ParamSpec, TypeVar

from loguru import logger


BASIC_SYSTEM_PROMPT = """\
You are a human preference judge tasked with deciding which of the two assistant responses, A or B, better responds to the user's prompt.

Respond with ONLY "A" if assistant A is better, "B" if assistant B is better, or "-" if neither is better than the other."""

USER_PROMPT_TEMPLATE = """\
<|Start of User Prompt|>
{prompt}
<|End of User Prompt|>

<|Start of Assistant A's Response|>
{response_a}
<|End of Assistant A's Response|>

<|Start of Assistant B's Response|>
{response_b}
<|End of Assistant B's Response|>"""

JOINED_PROMPT_TEMPLATE = """\
<|Start of System Prompt|>
{system_prompt}
<|End of System Prompt|>

{user_prompt}"""

ACCEPTABLE_RESPONSES = {"A", "B", "-"}


def get_user_prompt(prompt: str, response_a: str, response_b: str) -> str:
    return USER_PROMPT_TEMPLATE.format(prompt=prompt, response_a=response_a, response_b=response_b)


def rate_limit(
    *,
    n_calls: int,
    n_seconds: float,
    n_call_buffer: int = 50,
    max_wait_seconds: float = 60,
    backoff_seconds: float = 1,
) -> Callable:
    call_history: list[float] = []
    assert n_calls - n_call_buffer > 0, f"n_calls ({n_calls}) must be greater than n_call_buffer ({n_call_buffer})"

    def expire_old_calls() -> None:
        nonlocal call_history
        try:
            t_now = time.time()
            expiry_index = next((i for i, call_time in enumerate(call_history) if call_time < t_now - n_seconds))
            call_history = call_history[:expiry_index]
        except StopIteration:
            pass

    def can_call() -> bool:
        return len(call_history) < n_calls - n_call_buffer

    Params = ParamSpec("Params")
    ReturnType = TypeVar("ReturnType")

    def decorator(f: Callable[Params, ReturnType]) -> Callable[Params, ReturnType]:
        @functools.wraps(f)
        def wrapper(*args: Params.args, **kwargs: Params.kwargs) -> ReturnType:
            nonlocal call_history
            expire_old_calls()
            if not can_call():
                logger.warning(f"Hitting rate limit of {n_calls} calls per {n_seconds} seconds, waiting")
            for i in range(math.ceil(max_wait_seconds * backoff_seconds)):
                if not can_call():
                    time.sleep(backoff_seconds)
                expire_old_calls()
                if can_call():
                    break
                if i == max_wait_seconds - 1:
                    logger.error(f"Waited for {max_wait_seconds} seconds and still rate limited, exiting")
                    raise RuntimeError("rate limit exceeded")  # TODO: include name?
            call_history.insert(0, time.time())
            return f(*args, **kwargs)

        return wrapper

    return decorator



import time

import httpx
from loguru import logger
from openai import OpenAI
from pytimeparse.timeparse import timeparse

from autoarena.judge.base import AutomatedJudge
from autoarena.judge.utils import get_user_prompt, rate_limit


class OpenAIJudge(AutomatedJudge):
    API_KEY_NAME = "OPENAI_API_KEY"

    def __init__(self, name: str, model_name: str, system_prompt: str) -> None:
        super().__init__(name, model_name, system_prompt)
        self._client = OpenAI()

    @staticmethod
    def verify_environment() -> None:
        OpenAI().models.list()

    # OpenAI has different tiers and different rate limits for different models, choose a safeish value
    @rate_limit(n_calls=1_000, n_seconds=60)
    def judge(self, prompt: str, response_a: str, response_b: str) -> str:
        t0 = time.time()
        response_raw = self._client.chat.completions.with_raw_response.create(
            model=self.model_name,
            messages=[
                dict(role="system", content=self.system_prompt),
                dict(role="user", content=get_user_prompt(prompt, response_a, response_b)),
            ],
            max_tokens=self.MAX_TOKENS,
            timeout=httpx.Timeout(30),  # time out in 30 seconds
        )
        response = response_raw.parse()
        self.update_usage(response.usage.prompt_tokens, response.usage.completion_tokens, time.time() - t0)
        self._handle_rate_limit(dict(response_raw.headers))
        return response.choices[0].message.content

    # NOTE: ideally this would handle backoff for request rate limit, but the API surfaces the per-day, not the
    #  per-minute request rate limits and those aren't really actionable (hours until reset)
    @staticmethod
    def _handle_rate_limit(headers: dict[str, str]) -> None:
        try:
            request_limit = int(headers.get("x-ratelimit-remaining-requests", 1e6))
            if request_limit < 50:
                request_limit_reset = headers.get("x-ratelimit-reset-requests")
                logger.warning(
                    f"Approaching OpenAI request rate limit: {request_limit} remaining, resets in {request_limit_reset}"
                )
            token_limit = int(headers.get("x-ratelimit-remaining-tokens", 1e6))
            if token_limit < 5_000:
                token_limit_reset = headers.get("x-ratelimit-reset-tokens")
                logger.warning(
                    f"Approaching OpenAI token rate limit: {token_limit} remaining, backing off for {token_limit_reset}"
                )
                sleep_seconds = timeparse(token_limit_reset) or 0
                time.sleep(sleep_seconds)
        except Exception:  # don't let this crash the program as it's not clear that these headers are stable
            pass



import dataclasses
import time
from collections import defaultdict
from typing import Optional, Callable

import pandas as pd
from loguru import logger
from pydantic.dataclasses import dataclass
from tenacity import retry, stop_after_attempt, RetryCallState, wait_random_exponential

from autoarena.api import api
from autoarena.judge.base import AutomatedJudge
from autoarena.judge.executor import JudgeExecutor
from autoarena.judge.factory import judge_factory
from autoarena.judge.wrapper import JudgeWrapper, retrying_wrapper, fixing_wrapper, ab_shuffling_wrapper
from autoarena.service.elo import EloService, DEFAULT_ELO_CONFIG, EloConfig
from autoarena.service.head_to_head import HeadToHeadService
from autoarena.service.judge import JudgeService
from autoarena.service.model import ModelService
from autoarena.service.task import TaskService
from autoarena.store.utils import id_slug


class GracefulExit(RuntimeError): ...


@dataclass(frozen=True)
class AutoJudgeTask:
    project_slug: str
    task_id: int
    models: list[api.Model]
    judges: list[api.Judge]
    fraction: float = 1.0
    skip_existing: bool = False
    t_start: float = dataclasses.field(default_factory=time.time)
    judge_wrappers: list[JudgeWrapper] = dataclasses.field(
        default_factory=lambda: [retrying_wrapper, fixing_wrapper, ab_shuffling_wrapper]
    )
    update_every: int = 10
    elo_config: EloConfig = DEFAULT_ELO_CONFIG

    def __post_init__(self) -> None:
        if self.fraction <= 0 or self.fraction > 1:
            raise ValueError(f"Invalid fraction: {self.fraction} must be on (0,1]")
        if len(self.models) == 0:
            raise ValueError("No models provided")
        if len(self.judges) == 0:
            raise ValueError("No judges provided")

    @classmethod
    def create(
        cls,
        project_slug: str,
        models: Optional[list[api.Model]] = None,
        judges: Optional[list[api.Judge]] = None,
        fraction: float = 1.0,
        skip_existing: bool = False,
    ) -> Optional["AutoJudgeTask"]:
        models = models if models is not None else ModelService.get_all(project_slug)
        judges = judges if judges is not None else JudgeService.get_all(project_slug)
        enabled_judges = [j for j in judges if j.enabled and j.judge_type is not api.JudgeType.HUMAN]
        if len(enabled_judges) == 0:
            logger.warning("No enabled judges found, can't run automated judgement")
            return None  # do nothing if no judges are configured, do not create a task
        message = f"Started automated judging task using {len(enabled_judges)} judge(s)"
        task_id = TaskService.create(project_slug, api.TaskType.AUTO_JUDGE, message).id
        logger.info(message)
        return AutoJudgeTask(project_slug, task_id, models, enabled_judges, fraction, skip_existing)

    def log(
        self,
        message: str,
        status: api.TaskStatus = api.TaskStatus.IN_PROGRESS,
        progress: Optional[float] = None,
        level: str = "INFO",
    ) -> None:
        TaskService.update(self.project_slug, self.task_id, message, status=status, progress=progress)
        logger.log(level, message)

    # this is a beast, but most of the bulk comes from logging
    def run(self, executor: JudgeExecutor) -> None:
        try:
            self._run_inner(executor)
        except GracefulExit:
            pass
        except Exception as e:
            TaskService.update(self.project_slug, self.task_id, f"Failed ({e})", status=api.TaskStatus.FAILED)
            message = "See AutoArena service logs for more information"
            TaskService.update(self.project_slug, self.task_id, message, status=api.TaskStatus.FAILED)
            logger.error(f"Automated judgement failed: {e}")
            raise e

    def _retrieve_head_to_heads(self) -> pd.DataFrame:
        h2h_requests = [api.HeadToHeadsRequest(model_a_id=m.id) for m in self.models]
        df_h2h = pd.concat([HeadToHeadService.get_df(self.project_slug, request) for request in h2h_requests])
        if len(df_h2h) == 0:
            self.log("No head-to-heads found, exiting", status=api.TaskStatus.COMPLETED, progress=1, level="WARNING")
            raise GracefulExit

        df_h2h["response_id_slug"] = df_h2h.apply(lambda r: id_slug(r.response_a_id, r.response_b_id), axis=1)
        df_h2h = df_h2h.drop_duplicates(subset=["response_id_slug"], keep="first")
        n_models = len(set(df_h2h.model_a_id) | set(df_h2h.model_b_id))
        self.log(f"Found {len(df_h2h)} total head-to-heads between {n_models} model(s) to judge")

        if self.fraction < 1:
            n_total = len(df_h2h)
            df_h2h = df_h2h.sample(frac=self.fraction)
            self.log(f"Using subset of {len(df_h2h)} out of {n_total} head-to-heads ({int(100 * self.fraction)}%)")

        return df_h2h

    def _instantiate_judges_with_head_to_heads(
        self,
        df_h2h: pd.DataFrame,
    ) -> list[tuple[AutomatedJudge, list[api.HeadToHead]]]:
        judges_with_h2hs: list[tuple[AutomatedJudge, list[api.HeadToHead]]] = []
        for judge in self.judges:
            automated_judge = judge_factory(judge, wrappers=self.judge_wrappers)
            head_to_heads = [
                api.HeadToHead(r.prompt, r.response_a_id, r.response_a, r.response_b_id, r.response_b)
                for r in df_h2h.itertuples()
                if not self.skip_existing or judge.name not in {h["judge_name"] for h in r.history}
            ]
            n_skipping = len(df_h2h) - len(head_to_heads)
            if n_skipping > 0:
                self.log(f"Skipping {n_skipping} for '{judge.name}' with existing votes, {len(head_to_heads)} to run")
            if len(head_to_heads) == 0:
                message = f"No head-to-heads without votes found for '{judge.name}', skipping this judge"
                self.log(message, level="WARNING")
            else:
                judges_with_h2hs.append((automated_judge, head_to_heads))

        if len(judges_with_h2hs) == 0:
            message = "No head-to-heads without votes found for any judges, exiting"
            self.log(message, status=api.TaskStatus.COMPLETED, progress=1, level="WARNING")
            raise GracefulExit

        self.log(f"Running {len(judges_with_h2hs)} judge(s):")
        for j, _ in judges_with_h2hs:
            self.log(f"  * {j.name}")

        return judges_with_h2hs

    def _run_inner(self, executor: JudgeExecutor) -> None:
        df_h2h = self._retrieve_head_to_heads()
        judges_with_h2hs = self._instantiate_judges_with_head_to_heads(df_h2h)
        judge_id_by_name = {j.name: j.id for j in self.judges}
        out_columns = ["response_a_id", "response_b_id", "judge_id", "winner"]

        responses: dict[str, list[tuple[int, int, int, str]]] = defaultdict(lambda: [])
        n_h2h_by_judge_name = {judge.name: len(h2hs) for judge, h2hs in judges_with_h2hs}
        n_total = sum(len(h2hs) for _, h2hs in judges_with_h2hs)
        t_start_judging = time.time()
        for auto_judge, h2h, winner in executor.execute(judges_with_h2hs):
            judge_id = judge_id_by_name[auto_judge.name]
            responses[auto_judge.name].append((h2h.response_a_id, h2h.response_b_id, judge_id, winner))
            n_this_judge = len(responses[auto_judge.name])
            n_responses = sum(len(r) for r in responses.values())
            progress = 0.95 * (n_responses / n_total)
            if n_this_judge % self.update_every == 0:
                message = f"Judged {n_this_judge} of {n_h2h_by_judge_name[auto_judge.name]} with '{auto_judge.name}'"
                self.log(message, progress=progress)
                df_h2h_chunk = pd.DataFrame(responses[auto_judge.name][-self.update_every :], columns=out_columns)
                self._try_write(lambda: HeadToHeadService.upload_head_to_heads(self.project_slug, df_h2h_chunk))
            if n_this_judge == n_h2h_by_judge_name[auto_judge.name]:
                message = (
                    f"Judge '{auto_judge.name}' finished judging {n_h2h_by_judge_name[auto_judge.name]} head-to-heads "
                    f"in {time.time() - t_start_judging:0.1f} seconds"
                )
                self.log(message, progress=progress)
                for usage_summary_line in auto_judge.get_usage_summary():
                    self.log(usage_summary_line)
                df_h2h_all = pd.DataFrame(responses[auto_judge.name], columns=out_columns)
                self._try_write(lambda: HeadToHeadService.upload_head_to_heads(self.project_slug, df_h2h_all))

        self.log("Recomputing leaderboard rankings", progress=0.975)
        self._try_write(lambda: EloService.reseed_scores(self.project_slug, config=self.elo_config))
        message = f"Completed automated judging in {time.time() - self.t_start:0.1f} seconds"
        self.log(message, progress=1, status=api.TaskStatus.COMPLETED, level="SUCCESS")

    @staticmethod
    def _try_write(write_thunk: Callable[[], None]) -> None:
        """
        We're running into a limitation of DuckDB here -- concurrent writes are not well-supported. Give a best effort
        by retrying write attempts, an [officially endorsed strategy](https://duckdb.org/docs/connect/concurrency.html).

        This situation could be improved but requires answering tough questions about what kind of operations we want to
        optimize the persistence layer for, and what kind of operations we're OK with it struggling with.
        """

        def _log_retry(retry_state: RetryCallState) -> None:
            logger.warning(f"Retrying write attempt {retry_state.attempt_number} (error: {retry_state.outcome})")

        @retry(wait=wait_random_exponential(max=5), stop=stop_after_attempt(5), after=_log_retry)
        def _try_write_inner(thunk: Callable[[], None]) -> None:
            thunk()

        return _try_write_inner(write_thunk)



import inspect
import logging

from loguru import logger


class InterceptHandler(logging.Handler):
    """Intercept standard library log messages and emit via Loguru"""

    def emit(self, record: logging.LogRecord) -> None:
        """See https://loguru.readthedocs.io/en/stable/overview.html#entirely-compatible-with-standard-logging"""
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno
        frame, depth = inspect.currentframe(), 0
        while frame and (depth == 0 or frame.f_code.co_filename == logging.__file__):
            frame = frame.f_back
            depth += 1
        logger.opt(depth=depth, exception=record.exc_info).log(level, record.getMessage())


def initialize_logger() -> None:
    # replace uvicorn loggers with InterceptHandler
    intercept_handler = InterceptHandler()
    loggers = [logging.getLogger(name) for name in logging.root.manager.loggerDict if name.startswith("uvicorn.")]
    for uvicorn_logger in loggers:
        uvicorn_logger.handlers = []
    logging.getLogger("uvicorn").handlers = [intercept_handler]
    logging.getLogger("uvicorn.access").handlers = [intercept_handler]



from typing import Callable, TypeVar

import numpy as np
from loguru import logger
from tenacity import retry, stop_after_attempt, RetryCallState, wait_exponential

from autoarena.judge.base import AutomatedJudge
from autoarena.judge.utils import ACCEPTABLE_RESPONSES

JudgeWrapper = Callable[[type[AutomatedJudge]], type[AutomatedJudge]]
T = TypeVar("T", bound=AutomatedJudge)


def ab_shuffling_wrapper(judge_class: type[T]) -> type[T]:
    # not sure why mypy still complains about this after https://github.com/python/mypy/pull/14135
    class ABShufflingJudge(judge_class):  # type: ignore
        """Randomly shuffles which is A and which is B before passing to another judge."""

        def judge(self, prompt: str, response_a: str, response_b: str) -> str:
            shuffled = np.random.randint(2) == 0
            ra, rb = (response_b, response_a) if shuffled else (response_a, response_b)
            winner = super().judge(prompt, ra, rb)
            return self._unshuffle_winner(winner) if shuffled else winner

        @staticmethod
        def _unshuffle_winner(winner: str) -> str:
            return "B" if winner == "A" else "A" if winner == "B" else "-"

    return ABShufflingJudge


def clean_judgement(winner: str) -> str:
    return winner.strip(" \t\n'\"*.")  # strip common formatting issues


def cleaning_wrapper(judge_class: type[T]) -> type[T]:
    class CleaningJudge(judge_class):  # type: ignore
        """Attempt to clean raw responses from other judges"""

        def judge(self, prompt: str, response_a: str, response_b: str) -> str:
            winner_raw = super().judge(prompt, response_a, response_b)
            winner = clean_judgement(winner_raw)
            if winner in ACCEPTABLE_RESPONSES:
                return winner
            else:
                message = f"[{self.__class__.__name__}] Saving bad response from '{self.name}' as tie: {winner_raw}"
                logger.warning(message)
                return "-"

    return CleaningJudge


def fixing_wrapper(judge_class: type[T]) -> type[T]:
    class FixingJudge(judge_class):  # type: ignore
        """If the response does not fit nicely into the expected "A", "B", "-" format, classify it"""

        CLASSIFIER_MODEL = "MoritzLaurer/deberta-v3-xsmall-zeroshot-v1.1-all-33"
        A_IS_BETTER = "A is better"
        B_IS_BETTER = "B is better"
        TIE = "Neither is better / Both are good (Tie)"
        CLASSES = [A_IS_BETTER, B_IS_BETTER, TIE]
        CLASS_TO_WINNER = {A_IS_BETTER: "A", B_IS_BETTER: "B", TIE: "-"}

        def __init__(self, name: str, model_name: str, system_prompt: str):
            from transformers import pipeline

            super().__init__(name, model_name, system_prompt)
            self.pipe = pipeline(model=self.CLASSIFIER_MODEL, device="cpu")

        def judge(self, prompt: str, response_a: str, response_b: str) -> str:
            winner_raw = super().judge(prompt, response_a, response_b)
            winner = clean_judgement(winner_raw)
            if winner not in ACCEPTABLE_RESPONSES:
                classifications = self.pipe(winner_raw, candidate_labels=self.CLASSES)
                winner = self.CLASS_TO_WINNER[classifications["labels"][0]]
                logger.warning(f"Fixed bad response from '{self.name}': '{winner_raw}' as '{winner}'")
            return winner

    return FixingJudge


def retrying_wrapper(judge_class: type[T]) -> type[T]:
    class RetryingJudge(judge_class):  # type: ignore
        def judge(self, prompt: str, response_a: str, response_b: str) -> str:
            # wait 0, 1, 2, 4, 8, 16, 32 seconds
            @retry(wait=wait_exponential(min=0, max=32), stop=stop_after_attempt(7), after=self._log_retry)
            def judge_inner(p: str, ra: str, rb: str) -> str:
                return super(RetryingJudge, self).judge(p, ra, rb)

            return judge_inner(prompt, response_a, response_b)

        def _log_retry(self, retry_state: RetryCallState) -> None:
            message = f"Retrying '{self.name}' attempt {retry_state.attempt_number} (error: {retry_state.outcome})"
            logger.warning(message)

    return RetryingJudge



import dataclasses

import pandas as pd
from loguru import logger

from autoarena.api import api
from autoarena.error import BadRequestError
from autoarena.service.elo import EloService
from autoarena.service.project import ProjectService
from autoarena.store.utils import id_slug, check_required_columns


class HeadToHeadService:
    @staticmethod
    def get_df(project_slug: str, request: api.HeadToHeadsRequest) -> pd.DataFrame:
        with ProjectService.connect(project_slug) as conn:
            return conn.execute(
                """
                SELECT
                    ra.model_id AS model_a_id,
                    rb.model_id AS model_b_id,
                    ra.prompt AS prompt,
                    ra.id AS response_a_id,
                    rb.id AS response_b_id,
                    ra.response AS response_a,
                    rb.response AS response_b,
                    IFNULL(ARRAY_CONCAT(
                        ARRAY_AGG({
                            'judge_id': j1.id,
                            'judge_name': j1.name,
                            'winner': h1.winner,
                        }) FILTER (h1.winner IS NOT NULL),
                        ARRAY_AGG({
                            'judge_id': j2.id,
                            'judge_name': j2.name,
                            'winner': CASE WHEN h2.winner = 'A' THEN 'B'
                                           WHEN h2.winner = 'B' THEN 'A'
                                           ELSE h2.winner END,
                        }) FILTER (h2.winner IS NOT NULL)
                    ), []) AS history
                FROM response ra
                JOIN response rb ON ra.prompt = rb.prompt
                JOIN model ma ON ra.model_id = ma.id
                JOIN model mb ON rb.model_id = mb.id
                LEFT JOIN head_to_head h1 ON h1.response_a_id = ra.id AND h1.response_b_id = rb.id
                LEFT JOIN judge j1 ON j1.id = h1.judge_id
                LEFT JOIN head_to_head h2 ON h2.response_b_id = ra.id AND h2.response_a_id = rb.id
                LEFT JOIN judge j2 ON j2.id = h2.judge_id
                WHERE ra.model_id = $model_a_id
                AND ($model_b_id IS NULL OR rb.model_id = $model_b_id)
                AND ra.model_id != rb.model_id
                GROUP BY ra.model_id, rb.model_id, ra.id, rb.id, ra.prompt, ra.response, rb.response
                ORDER BY ra.id, rb.id
                """,
                dict(model_a_id=request.model_a_id, model_b_id=request.model_b_id),
            ).df()

    @staticmethod
    def get(project_slug: str, request: api.HeadToHeadsRequest) -> list[api.HeadToHead]:
        df_h2h = HeadToHeadService.get_df(project_slug, request)
        return [
            api.HeadToHead(
                prompt=r.prompt,
                response_a_id=r.response_a_id,
                response_a=r.response_a,
                response_b_id=r.response_b_id,
                response_b=r.response_b,
                history=[api.HeadToHeadHistoryItem(**h) for h in r.history],
            )
            for r in df_h2h.itertuples()
        ]

    @staticmethod
    def get_count(project_slug: str) -> int:
        with ProjectService.connect(project_slug) as conn:
            ((n_h2h,),) = conn.execute(
                """
                SELECT COUNT(DISTINCT id_slug(ra.id, rb.id))
                FROM response ra
                JOIN response rb ON ra.prompt = rb.prompt AND ra.model_id != rb.model_id
                """,
            ).fetchall()
        return n_h2h

    @staticmethod
    def submit_vote(project_slug: str, request: api.HeadToHeadVoteRequest) -> None:
        with ProjectService.connect(project_slug) as conn:
            # 1. insert head-to-head record
            conn.execute(
                """
                INSERT INTO head_to_head (response_id_slug, response_a_id, response_b_id, judge_id, winner)
                SELECT ID_SLUG($response_a_id, $response_b_id), $response_a_id, $response_b_id, j.id, $winner
                FROM judge j
                WHERE j.judge_type = $judge_type
                ON CONFLICT (response_id_slug, judge_id) DO UPDATE SET
                    winner = IF(response_a_id = $response_b_id, INVERT_WINNER(EXCLUDED.winner), EXCLUDED.winner)
            """,
                dict(**dataclasses.asdict(request), judge_type=api.JudgeType.HUMAN.value),
            )

            # 2. adjust elo scores
            df_model = conn.execute(
                """
                SELECT id, elo
                FROM model m
                WHERE EXISTS (SELECT 1 FROM response r WHERE r.id = $response_a_id AND r.model_id = m.id)
                UNION ALL
                SELECT id, elo
                FROM model m
                WHERE EXISTS (SELECT 1 FROM response r WHERE r.id = $response_b_id AND r.model_id = m.id)
            """,
                dict(response_a_id=request.response_a_id, response_b_id=request.response_b_id),
            ).df()
            model_a = df_model.iloc[0]
            model_b = df_model.iloc[1]
            elo_a, elo_b = EloService.compute_elo_single(model_a.elo, model_b.elo, request.winner)
            for model_id, elo in [(model_a.id, elo_a), (model_b.id, elo_b)]:
                conn.execute("UPDATE model SET elo = $elo WHERE id = $model_id", dict(model_id=model_id, elo=elo))

    @staticmethod
    def upload_head_to_heads(project_slug: str, df_h2h: pd.DataFrame) -> None:  # TODO: return type?
        try:
            check_required_columns(df_h2h, ["response_a_id", "response_b_id", "judge_id", "winner"])
        except ValueError as e:
            raise BadRequestError(str(e))
        df_h2h_deduped = df_h2h.copy()
        df_h2h_deduped["response_id_slug"] = df_h2h_deduped.apply(
            lambda r: id_slug(r.response_a_id, r.response_b_id), axis=1
        )
        df_h2h_deduped = df_h2h_deduped.drop_duplicates(subset=["response_id_slug", "judge_id"], keep="first")
        if len(df_h2h_deduped) != len(df_h2h):
            logger.warning(f"Dropped {len(df_h2h) - len(df_h2h_deduped)} duplicate rows before uploading")
        with ProjectService.connect(project_slug) as conn:
            conn.execute("""
                INSERT INTO head_to_head (response_id_slug, response_a_id, response_b_id, judge_id, winner)
                SELECT id_slug(response_a_id, response_b_id), response_a_id, response_b_id, judge_id, winner
                FROM df_h2h_deduped
                ON CONFLICT (response_id_slug, judge_id) DO UPDATE SET winner = EXCLUDED.winner
            """)



import argparse
from pathlib import Path

import uvicorn

from autoarena.seed import seed_head_to_heads


def parse_args(args: list[str]) -> argparse.Namespace:
    ap = argparse.ArgumentParser("autoarena")
    sp = ap.add_subparsers(dest="command")
    sp.default = "serve"

    serve_parser = sp.add_parser("serve", help="[Default] Serve the AutoArena app")
    serve_parser.add_argument("-d", "--dev", action="store_true", help="Run in development mode")

    seed_parser = sp.add_parser(
        "seed",
        help="Seed a project with head-to-heads stored in a CSV or Parquet file",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="""\
Seed a new project from a CSV or Parquet file where each row represents a head-to-head matchup between two models.

The following columns are required:

- `model_a`: name of model A in this head-to-head
- `model_b`: name of model B in this head-to-head
- `prompt`: the prompt that both models were run on
- `response_a`: the response from model A to the prompt
- `response_b`: the response from model B to the prompt
- `winner`: the winner of the head-to-head, either "A", "B", or "-" for ties""",
    )
    seed_parser.add_argument("head_to_heads", type=Path, help="Path to head-to-heads CSV or Parquet file")

    return ap.parse_args(args)


def main(args: list[str]) -> None:
    parsed_args = parse_args(args)
    if parsed_args.command == "seed":
        seed_head_to_heads(parsed_args.head_to_heads)
    if parsed_args.command == "serve":
        uvicorn.run(
            "autoarena.server:server",
            host="localhost",
            port=8899,
            reload=getattr(parsed_args, "dev", False),
            factory=True,
            timeout_graceful_shutdown=1,  # wait 1 second for tasks to complete before forcefully exiting
        )



import asyncio
import time
from datetime import datetime
from typing import Optional, AsyncIterator


from autoarena.api import api
from autoarena.error import NotFoundError
from autoarena.judge.executor import ThreadedExecutor
from autoarena.service.elo import EloService
from autoarena.service.project import ProjectService


class TaskService:
    @staticmethod
    def get_all(project_slug: str) -> list[api.Task]:
        with ProjectService.connect(project_slug) as conn:
            df_task = conn.execute("SELECT id, task_type, created, progress, status, logs FROM task").df()
        return [api.Task(**r) for _, r in df_task.iterrows()]

    @staticmethod
    def get(project_slug: str, task_id: int) -> api.Task:
        try:
            with ProjectService.connect(project_slug) as conn:
                df_task = conn.execute(
                    "SELECT id, task_type, created, progress, status, logs FROM task WHERE id = $task_id",
                    dict(task_id=task_id),
                ).df()
                return [api.Task(**r) for _, r in df_task.iterrows()][0]
        except IndexError:
            raise NotFoundError(f"Task with id '{task_id}' not found")

    @staticmethod
    async def get_stream(project_slug: str, task_id: int) -> AsyncIterator[api.Task]:
        prev: Optional[api.Task] = None
        while True:
            cur = TaskService.get(project_slug, task_id)
            if prev != cur:
                yield cur
            if cur.status in {api.TaskStatus.COMPLETED, api.TaskStatus.FAILED}:
                break
            prev = cur
            await asyncio.sleep(0.2)

    @staticmethod
    def has_active(project_slug: str) -> api.HasActiveTasks:
        with ProjectService.connect(project_slug) as conn:
            records = conn.execute(
                "SELECT 1 WHERE EXISTS (SELECT 1 FROM task WHERE status IN ($started, $in_progress))",
                dict(started=api.TaskStatus.STARTED.value, in_progress=api.TaskStatus.IN_PROGRESS.value),
            ).fetchall()
            return api.HasActiveTasks(has_active=len(records) > 0)

    @staticmethod
    async def has_active_stream(
        project_slug: str,
        timeout: Optional[float] = None,
    ) -> AsyncIterator[api.HasActiveTasks]:
        t0 = time.time()
        prev: Optional[api.HasActiveTasks] = None
        while timeout is None or time.time() - t0 < timeout:
            cur = TaskService.has_active(project_slug)
            if prev != cur:
                yield cur
            prev = cur
            await asyncio.sleep(1)

    @staticmethod
    def create(project_slug: str, task_type: api.TaskType, log: str = "Started") -> api.Task:
        with ProjectService.connect(project_slug) as conn:
            logs = f"{TaskService._time_slug()} {log}"
            ((task_id, created, progress, status, logs),) = conn.execute(
                """
                INSERT INTO task (task_type, status, logs)
                VALUES ($task_type, $status, $logs)
                RETURNING id, created, progress, status, logs
                """,
                dict(task_type=task_type.value, status=api.TaskStatus.STARTED.value, logs=logs),
            ).fetchall()
        return api.Task(id=task_id, task_type=task_type, created=created, progress=progress, status=status, logs=logs)

    @staticmethod
    def delete_completed(project_slug: str) -> None:
        with ProjectService.connect(project_slug) as conn:
            conn.execute("TRUNCATE task")

    @staticmethod
    def update(
        project_slug: str,
        task_id: int,
        log: str,
        progress: Optional[float] = None,
        status: api.TaskStatus = api.TaskStatus.IN_PROGRESS,
    ) -> None:
        with ProjectService.connect(project_slug) as conn:
            log = f"{TaskService._time_slug()} {log}"
            conn.execute(
                """
                UPDATE task
                SET progress = IFNULL($progress, progress),
                    status = $status,
                    logs = logs || '\n' || $log
                WHERE id = $id
                """,
                dict(id=task_id, log=log, progress=progress, status=status.value),
            )

    @staticmethod
    def _time_slug() -> str:
        return datetime.now().strftime("[%Y-%m-%d %H:%M:%S]")

    # TODO: should this really be a long-running task? It only takes ~5 seconds for ~50k head-to-heads
    @staticmethod
    def recompute_leaderboard(project_slug: str) -> None:
        task_objects = TaskService.get_all(project_slug)
        if len([t for t in task_objects if t.task_type is api.TaskType.RECOMPUTE_LEADERBOARD and t.progress < 1]) > 0:
            return  # only recompute if there isn't already a task in progress
        task_id = TaskService.create(project_slug, api.TaskType.RECOMPUTE_LEADERBOARD).id
        try:
            EloService.reseed_scores(project_slug)
        finally:
            TaskService.update(project_slug, task_id, "Done", progress=1, status=api.TaskStatus.COMPLETED)

    @staticmethod
    def auto_judge(
        project_slug: str,
        *,
        models: Optional[list[api.Model]] = None,
        judges: Optional[list[api.Judge]] = None,
        fraction: float = 1.0,
        skip_existing: bool = False,
    ) -> None:
        from autoarena.task.auto_judge import AutoJudgeTask

        auto_judge_task = AutoJudgeTask.create(project_slug, models, judges, fraction, skip_existing)
        if auto_judge_task is not None:
            with ThreadedExecutor(8) as executor:
                auto_judge_task.run(executor)



from typing import TypeVar, AsyncIterator

from pydantic import RootModel
from starlette.responses import StreamingResponse

T = TypeVar("T")  # should be a Pydantic dataclass


async def as_sse_stream(object_stream: AsyncIterator[T]) -> AsyncIterator[str]:
    async for obj in object_stream:
        obj_json = RootModel[T](obj).model_dump_json()
        yield f"data: {obj_json}\n\n"


# TODO: figure out a way to gracefully terminate this when the server is exiting
class SSEStreamingResponse(StreamingResponse):
    def __init__(self, object_stream: AsyncIterator[T]):
        super().__init__(as_sse_stream(object_stream), media_type="text/event-stream")



import dataclasses
from datetime import datetime
from enum import Enum
from typing import Literal, Optional, Union

from pydantic.dataclasses import dataclass


@dataclass(frozen=True)
class Project:
    slug: str  # stem of the database file
    filename: str  # name of database file
    filepath: str  # full path to database file


@dataclass(frozen=True)
class CreateProjectRequest:
    name: str


@dataclass(frozen=True)
class Model:
    id: int
    name: str
    created: datetime
    elo: float
    q025: Optional[float]
    q975: Optional[float]
    n_responses: int
    n_votes: int


@dataclass(frozen=True)
class ModelResponse:
    prompt: str
    response: str


@dataclass(frozen=True)
class ModelHeadToHeadStats:
    other_model_id: int
    other_model_name: str
    judge_id: int
    judge_name: str
    count_wins: int
    count_losses: int
    count_ties: int


@dataclass(frozen=True)
class HeadToHeadsRequest:
    model_a_id: int
    model_b_id: Optional[int] = None  # when empty, get all pairings


WinnerType = Union[Literal["A", "B", "-"], str]  # should be one of the literal values, but could be anything


@dataclass(frozen=True)
class HeadToHeadHistoryItem:
    judge_id: int
    judge_name: str
    winner: WinnerType


@dataclass(frozen=True)
class HeadToHead:
    prompt: str
    response_a_id: int
    response_a: str
    response_b_id: int
    response_b: str
    history: list[HeadToHeadHistoryItem] = dataclasses.field(default_factory=list)


@dataclass(frozen=True)
class HeadToHeadVoteRequest:  # this is always coming from humans
    response_a_id: int
    response_b_id: int
    winner: WinnerType


class TaskType(str, Enum):
    AUTO_JUDGE = "auto-judge"
    RECOMPUTE_LEADERBOARD = "recompute-leaderboard"
    FINE_TUNE = "fine-tune"


class TaskStatus(str, Enum):
    STARTED = "started"
    IN_PROGRESS = "in-progress"
    COMPLETED = "completed"
    FAILED = "failed"


@dataclass(frozen=True)
class Task:
    id: int
    task_type: TaskType
    created: datetime
    progress: float  # on [0,1]
    status: TaskStatus
    logs: str


@dataclass(frozen=True)
class HasActiveTasks:
    has_active: bool


@dataclass(frozen=True)
class TriggerAutoJudgeRequest:
    judge_ids: list[int]
    fraction: float  # on [0,1]
    skip_existing: bool


class JudgeType(str, Enum):
    HUMAN = "human"
    OLLAMA = "ollama"
    OPENAI = "openai"
    ANTHROPIC = "anthropic"
    COHERE = "cohere"
    GEMINI = "gemini"
    TOGETHER = "together"
    BEDROCK = "bedrock"
    CUSTOM = "custom"  # TODO: not sure how to handle calling this yet -- will it just be another Ollama model?


@dataclass(frozen=True)
class Judge:
    id: int
    judge_type: JudgeType
    created: datetime
    name: str
    model_name: Optional[str]
    system_prompt: Optional[str]
    description: str
    enabled: bool
    n_votes: int


@dataclass(frozen=True)
class CreateJudgeRequest:
    judge_type: JudgeType
    name: str
    model_name: str
    system_prompt: str
    description: str


@dataclass(frozen=True)
class UpdateJudgeRequest:
    enabled: bool
    # TODO: update name, description, system prompt?


@dataclass(frozen=True)
class EloHistoryItem:
    other_model_id: int
    other_model_name: str
    judge_id: int
    judge_name: str
    elo: float


@dataclass(frozen=True)
class CreateFineTuningTaskRequest:
    base_model: str



import time

import anthropic

from autoarena.judge.base import AutomatedJudge
from autoarena.judge.utils import get_user_prompt, rate_limit


class AnthropicJudge(AutomatedJudge):
    API_KEY_NAME = "ANTHROPIC_API_KEY"

    def __init__(self, name: str, model_name: str, system_prompt: str) -> None:
        super().__init__(name, model_name, system_prompt)
        self._client = anthropic.Client()

    @staticmethod
    def verify_environment() -> None:
        # this is a little dirty, but gets the job done as requests without valid auth are rejected eagerly
        try:
            anthropic.Client().post("/v1/messages", cast_to=object)
        except (anthropic.AuthenticationError, TypeError) as e:
            raise e
        except Exception:
            pass

    # anthropic has different tiers with 1000/2000/4000, opting to be conservative by default
    @rate_limit(n_calls=1_000, n_seconds=60)
    def judge(self, prompt: str, response_a: str, response_b: str) -> str:
        t0 = time.time()
        response = self._client.messages.create(
            model=self.model_name,
            system=self.system_prompt,
            messages=[dict(role="user", content=get_user_prompt(prompt, response_a, response_b))],
            max_tokens=self.MAX_TOKENS,
        )
        self.update_usage(response.usage.input_tokens, response.usage.output_tokens, time.time() - t0)
        return response.content[0].text



from pathlib import Path

import pandas as pd

from autoarena.api import api
from autoarena.service.elo import EloService
from autoarena.service.head_to_head import HeadToHeadService
from autoarena.service.judge import JudgeService
from autoarena.service.model import ModelService
from autoarena.service.project import ProjectService
from autoarena.store.utils import check_required_columns


def seed_head_to_heads(head_to_heads: Path) -> None:
    df = pd.read_parquet(head_to_heads) if head_to_heads.suffix == ".parquet" else pd.read_csv(head_to_heads)
    check_required_columns(df, ["model_a", "model_b", "prompt", "response_a", "response_b", "winner"])

    # 1. seed project
    project_slug = ProjectService.create_idempotent(api.CreateProjectRequest(name=head_to_heads.stem)).slug

    # 2. seed models
    models = set(df.model_a) | set(df.model_b)
    model_ids: list[int] = []
    for model in models:
        cols = ["prompt", "response"]
        df_model_response_a = df[df.model_a == model].rename(columns=dict(response_a="response"))[cols]
        df_model_response_b = df[df.model_b == model].rename(columns=dict(response_b="response"))[cols]
        df_model_response = pd.concat([df_model_response_a, df_model_response_b])
        df_model_response = df_model_response.drop_duplicates(subset=["prompt"], keep="last")  # drop duplicate rows
        df_model_response = df_model_response.dropna(subset=["response"])
        model_ids.append(ModelService.upload_responses(project_slug, model, df_model_response).id)

    # 3. seed head-to-heads
    df_response = pd.concat([ModelService.get_df_response(project_slug, model_id) for model_id in model_ids])
    right_on = ["model", "prompt", "response"]
    df = df.merge(df_response, left_on=["model_a", "prompt", "response_a"], right_on=right_on, how="left")
    df = df.rename(columns=dict(response_id="response_a_id"))
    df = df.merge(df_response, left_on=["model_b", "prompt", "response_b"], right_on=right_on, how="left")
    df = df.rename(columns=dict(response_id="response_b_id"))
    df = df.dropna(subset=["response_a_id", "response_b_id"])
    df[["response_a_id", "response_b_id"]] = df[["response_a_id", "response_b_id"]].astype(int)
    # TODO: allow seeding with non-human judgements?
    df["judge_id"] = [j for j in JudgeService.get_all(project_slug) if j.judge_type is api.JudgeType.HUMAN][0].id
    HeadToHeadService.upload_head_to_heads(project_slug, df[["response_a_id", "response_b_id", "judge_id", "winner"]])

    # 4. seed elo scores
    EloService.reseed_scores(project_slug)



from typing import Optional, Sequence

from autoarena.api import api
from autoarena.judge.anthropic import AnthropicJudge
from autoarena.judge.base import AutomatedJudge
from autoarena.judge.bedrock import BedrockJudge
from autoarena.judge.cohere import CohereJudge
from autoarena.judge.custom import get_custom_judge_class
from autoarena.judge.gemini import GeminiJudge
from autoarena.judge.ollama import OllamaJudge
from autoarena.judge.openai import OpenAIJudge
from autoarena.judge.together import TogetherJudge
from autoarena.judge.wrapper import JudgeWrapper

AUTOMATED_JUDGE_TYPE_TO_CLASS: dict[api.JudgeType, type[AutomatedJudge]] = {
    api.JudgeType.OLLAMA: OllamaJudge,
    api.JudgeType.OPENAI: OpenAIJudge,
    api.JudgeType.ANTHROPIC: AnthropicJudge,
    api.JudgeType.COHERE: CohereJudge,
    api.JudgeType.GEMINI: GeminiJudge,
    api.JudgeType.TOGETHER: TogetherJudge,
    api.JudgeType.BEDROCK: BedrockJudge,
}


def judge_factory(judge: api.Judge, wrappers: Optional[Sequence[JudgeWrapper]] = None) -> AutomatedJudge:
    if judge.judge_type is api.JudgeType.HUMAN:
        raise ValueError("automated judge factory cannot instantiate human judge")
    judge_class = (
        get_custom_judge_class(judge.name)
        if judge.judge_type is api.JudgeType.CUSTOM
        else AUTOMATED_JUDGE_TYPE_TO_CLASS[judge.judge_type]
    )
    if not issubclass(judge_class, AutomatedJudge) or judge.model_name is None or judge.system_prompt is None:
        raise ValueError(f"misconfigured judge: {judge}")
    for wrapper in wrappers or []:
        judge_class = wrapper(judge_class)
    return judge_class(judge.name, judge.model_name, judge.system_prompt)


def verify_judge_type_environment(judge_type: api.JudgeType) -> None:
    judge_class = AUTOMATED_JUDGE_TYPE_TO_CLASS.get(judge_type, None)
    if judge_class is not None:
        judge_class.verify_environment()



from pathlib import Path

from fastapi import APIRouter
from fastapi.responses import PlainTextResponse
from starlette.responses import FileResponse, HTMLResponse

from autoarena.error import NotFoundError

ROBOTS_TXT = "User-agent: *\nDisallow: /"


def ui_router() -> APIRouter:
    build_static_dir = Path(__file__).parent / "ui"
    dev_static_dir = Path(__file__).parent.parent / "ui" / "dist"
    static_dir = build_static_dir if build_static_dir.exists() else dev_static_dir

    r = APIRouter()

    @r.get("/robots.txt", response_class=PlainTextResponse)
    async def robots_txt() -> str:
        return ROBOTS_TXT

    @r.get("/assets/{file_path:path}")
    async def assets(file_path: str) -> FileResponse:
        path = static_dir / "assets" / file_path
        if not path.is_file():
            raise NotFoundError("file not found")
        return FileResponse(path)

    @r.get("/{full_path:path}")
    async def catch_all(full_path: str) -> HTMLResponse:
        with (static_dir / "index.html").open("r") as f:
            return HTMLResponse(f.read())

    return r



from loguru import logger

from autoarena.api import api
from autoarena.judge.factory import verify_judge_type_environment
from autoarena.judge.utils import BASIC_SYSTEM_PROMPT
from autoarena.service.project import ProjectService


class JudgeService:
    @staticmethod
    def get_default_system_prompt() -> str:
        return BASIC_SYSTEM_PROMPT

    @staticmethod
    def get_all(project_slug: str) -> list[api.Judge]:
        with ProjectService.connect(project_slug) as conn:
            df_task = conn.execute(
                """
                SELECT
                    j.id,
                    j.judge_type,
                    j.created,
                    j.name,
                    j.model_name,
                    j.system_prompt,
                    j,description,
                    j.enabled,
                    SUM(IF(h.id IS NOT NULL, 1, 0)) AS n_votes
                FROM judge j
                LEFT JOIN head_to_head h ON h.judge_id = j.id
                GROUP BY
                    j.id,
                    j.judge_type,
                    j.created,
                    j.name,
                    j.model_name,
                    j.system_prompt,
                    j.description,
                    j.enabled
                ORDER BY j.id
                """,
            ).df()
        return [api.Judge(**r) for _, r in df_task.iterrows()]

    @staticmethod
    def create(project_slug: str, request: api.CreateJudgeRequest) -> api.Judge:
        with ProjectService.connect(project_slug) as conn:
            ((judge_id, created, enabled),) = conn.execute(
                """
                INSERT INTO judge (judge_type, name, model_name, system_prompt, description, enabled)
                VALUES ($judge_type, $name, $model_name, $system_prompt, $description, TRUE)
                RETURNING id, created, enabled
            """,
                dict(
                    judge_type=request.judge_type.value,
                    name=request.name,
                    model_name=request.model_name,
                    system_prompt=request.system_prompt,
                    description=request.description,
                ),
            ).fetchall()
        return api.Judge(
            id=judge_id,
            judge_type=request.judge_type,
            created=created,
            name=request.name,
            model_name=request.model_name,
            system_prompt=request.system_prompt,
            description=request.description,
            enabled=enabled,
            n_votes=0,
        )

    # TODO: is this necessary?
    @staticmethod
    def create_human_judge(project_slug: str) -> api.Judge:
        with ProjectService.connect(project_slug) as conn:
            conn.execute(
                """
                INSERT INTO judge (judge_type, name, description, enabled)
                VALUES ($judge_type, $name, $description, TRUE)
                ON CONFLICT (name) DO NOTHING
            """,
                dict(
                    judge_type=api.JudgeType.HUMAN.value,
                    name=api.JudgeType.HUMAN.value,
                    description="Manual ratings submitted via the 'Head-to-Head' tab",
                ),
            )
        # TODO: this is a little lazy but ¯\_(ツ)_/¯
        return [j for j in JudgeService.get_all(project_slug) if j.judge_type is api.JudgeType.HUMAN][0]

    @staticmethod
    def update(project_slug: str, judge_id: int, request: api.UpdateJudgeRequest) -> api.Judge:
        with ProjectService.connect(project_slug) as conn:
            conn.execute(
                "UPDATE judge SET enabled = $enabled WHERE id = $judge_id",
                dict(judge_id=judge_id, enabled=request.enabled),
            )
        return [j for j in JudgeService.get_all(project_slug) if j.id == judge_id][0]

    @staticmethod
    def delete(project_slug: str, judge_id: int) -> None:
        with ProjectService.connect(project_slug) as conn:
            conn.execute("DELETE FROM head_to_head WHERE judge_id = $judge_id", dict(judge_id=judge_id))
            conn.execute("DELETE FROM judge WHERE id = $judge_id", dict(judge_id=judge_id))

    @staticmethod
    def check_can_access(judge_type: api.JudgeType) -> bool:
        try:
            verify_judge_type_environment(judge_type)
            return True
        except Exception as e:
            logger.warning(f"Missing environment configuration necessary to run '{judge_type}' judge: {e}")
            return False



import time

import boto3

from autoarena.judge.base import AutomatedJudge
from autoarena.judge.utils import rate_limit, get_user_prompt


class BedrockJudge(AutomatedJudge):
    API_KEY_NAME = None

    def __init__(self, name: str, model_name: str, system_prompt: str) -> None:
        super().__init__(name, model_name, system_prompt)
        self._client = boto3.client(service_name="bedrock-runtime")

    @staticmethod
    def verify_environment() -> None:
        boto3.client(service_name="sts").get_caller_identity()

    @rate_limit(n_calls=200, n_seconds=1, n_call_buffer=25)
    def judge(self, prompt: str, response_a: str, response_b: str) -> str:
        t0 = time.time()
        response = self._client.converse(
            modelId=self.model_name,
            system=[dict(text=self.system_prompt)],
            messages=[dict(role="user", content=[dict(text=get_user_prompt(prompt, response_a, response_b))])],
            inferenceConfig=dict(maxTokens=self.MAX_TOKENS),
        )
        self.update_usage(response["usage"]["inputTokens"], response["usage"]["outputTokens"], time.time() - t0)
        return response["output"]["message"]["content"][0]["text"]



import numpy as np
import pandas as pd
from loguru import logger

from autoarena.api import api
from autoarena.error import NotFoundError, BadRequestError
from autoarena.service.elo import EloService, DEFAULT_ELO_CONFIG
from autoarena.service.project import ProjectService
from autoarena.store.utils import check_required_columns


class ModelService:
    MODELS_QUERY = """
        WITH response_count AS (
            SELECT r.model_id, COUNT(1) AS response_count
            FROM response r
            GROUP BY r.model_id
        ), vote_count_a AS ( -- this is inelegant but this query is tricky to write
            SELECT m.id AS model_id, SUM(IF(h.id IS NOT NULL, 1, 0)) AS vote_count
            FROM model m
            JOIN response r ON r.model_id = m.id
            LEFT JOIN head_to_head h ON r.id = h.response_a_id
            GROUP BY m.id
        ), vote_count_b AS (
            SELECT m.id AS model_id, SUM(IF(h.id IS NOT NULL, 1, 0)) AS vote_count
            FROM model m
            JOIN response r ON r.model_id = m.id
            LEFT JOIN head_to_head h ON r.id = h.response_b_id
            GROUP BY m.id
        )
        SELECT
            id,
            name,
            created,
            elo,
            q025,
            q975,
            IFNULL(rc.response_count, 0) AS n_responses,
            IFNULL(vca.vote_count, 0) + IFNULL(vcb.vote_count, 0) AS n_votes
        FROM model m
        LEFT JOIN response_count rc ON m.id = rc.model_id
        LEFT JOIN vote_count_a vca ON m.id = vca.model_id
        LEFT JOIN vote_count_b vcb ON m.id = vcb.model_id
        """

    @staticmethod
    def get_by_id(project_slug: str, model_id: int) -> api.Model:
        with ProjectService.connect(project_slug) as conn:
            df_model = conn.execute(f"{ModelService.MODELS_QUERY} WHERE m.id = $model_id", dict(model_id=model_id)).df()
        try:
            return [api.Model(**r) for _, r in df_model.iterrows()][0]
        except IndexError:
            raise NotFoundError(f"Model with ID '{model_id}' not found")

    @staticmethod
    def get_all_df(project_slug: str) -> pd.DataFrame:
        with ProjectService.connect(project_slug) as conn:
            df_model = conn.execute(ModelService.MODELS_QUERY).df()
        df_model = df_model.replace({np.nan: None})
        return df_model

    @staticmethod
    def get_all(project_slug: str) -> list[api.Model]:
        df_model = ModelService.get_all_df(project_slug)
        return [api.Model(**r) for _, r in df_model.iterrows()]

    @staticmethod
    def get_all_ranked_by_judge(project_slug: str, judge_id: int) -> list[api.Model]:
        df_h2h = EloService.get_df_head_to_head(project_slug)
        df_h2h = df_h2h[df_h2h["judge_id"] == judge_id]
        df_elo = EloService.compute_elo(df_h2h)
        df_model = ModelService.get_all_df(project_slug)
        df_out = pd.merge(df_model, df_elo, left_on="name", right_on="model", how="left")
        df_out[["elo", "q025", "q975"]] = df_out[["elo_y", "q025_y", "q975_y"]]
        df_out["elo"] = df_out["elo"].replace({np.nan: DEFAULT_ELO_CONFIG.default_score})
        df_out = df_out.replace({np.nan: None})
        votes_a, votes_b = df_h2h.model_a_id.value_counts(), df_h2h.model_b_id.value_counts()
        df_votes = pd.merge(votes_a, votes_b, left_index=True, right_index=True, how="outer")
        df_votes = df_votes.replace({np.nan: 0})
        df_votes["n_votes"] = df_votes["count_x"] + df_votes["count_y"]
        df_out = df_out.merge(df_votes, left_on="id", right_index=True, how="left")
        df_out["n_votes"] = df_out["n_votes_y"].replace({np.nan: 0})
        df_out = df_out[["id", "name", "created", "elo", "q025", "q975", "n_responses", "n_votes"]]
        return [api.Model(**r) for _, r in df_out.iterrows()]

    @staticmethod
    def upload_responses(project_slug: str, model_name: str, df_response: pd.DataFrame) -> api.Model:
        try:
            check_required_columns(df_response, ["prompt", "response"])
        except ValueError as e:
            raise BadRequestError(str(e))
        n_input = len(df_response)
        df_response = df_response.copy().dropna(subset=["prompt", "response"])
        if len(df_response) != n_input:
            logger.warning(f"Dropped {n_input - len(df_response)} responses with empty prompt or response values")
        logger.info(f"Uploading {len(df_response)} responses from model '{model_name}'")
        with ProjectService.connect(project_slug) as conn:
            ((new_model_id,),) = conn.execute(
                "INSERT INTO model (name) VALUES ($model_name) RETURNING id",
                dict(model_name=model_name),
            ).fetchall()
            df_response["model_id"] = new_model_id
            conn.execute("""
                INSERT INTO response (model_id, prompt, response)
                SELECT model_id, prompt, response
                FROM df_response
            """)
        models = ModelService.get_all(project_slug)
        new_model = [model for model in models if model.id == new_model_id][0]
        return new_model

    @staticmethod
    def delete(project_slug: str, model_id: int) -> None:
        params = dict(model_id=model_id)
        with ProjectService.connect(project_slug) as conn:
            conn.execute(
                """
                DELETE FROM head_to_head h
                WHERE EXISTS (
                    SELECT 1
                    FROM response r
                    WHERE r.model_id = $model_id
                    AND (h.response_a_id = r.id OR h.response_b_id = r.id)
                )
                """,
                params,
            )
            conn.execute("DELETE FROM response WHERE model_id = $model_id", params)
            conn.execute("DELETE FROM model WHERE id = $model_id", params)

    @staticmethod
    def get_responses(project_slug: str, model_id: int) -> list[api.ModelResponse]:
        df_response = ModelService.get_df_response(project_slug, model_id)
        return [api.ModelResponse(prompt=r.prompt, response=r.response) for r in df_response.itertuples()]

    @staticmethod
    def get_df_response(project_slug: str, model_id: int) -> pd.DataFrame:
        with ProjectService.connect(project_slug) as conn:
            df_response = conn.execute(
                """
                SELECT
                    m.name AS model,
                    r.id AS response_id,
                    r.prompt AS prompt,
                    r.response AS response
                FROM model m
                JOIN response r ON m.id = r.model_id
                WHERE m.id = $model_id
            """,
                dict(model_id=model_id),
            ).df()
        if len(df_response) == 0:
            raise NotFoundError(f"Model with ID '{model_id}' not found")  # model can't exist without any responses
        return df_response

    @staticmethod
    def get_df_head_to_head(project_slug: str, model_id: int) -> pd.DataFrame:
        with ProjectService.connect(project_slug) as conn:
            df_h2h = conn.execute(
                """
                SELECT
                    ra.prompt,
                    ma.name AS model_a,
                    mb.name AS model_b,
                    ra.response AS response_a,
                    rb.response AS response_b,
                    j.name AS judge,
                    h.winner
                FROM head_to_head h
                JOIN judge j ON h.judge_id = j.id
                JOIN response ra ON ra.id = h.response_a_id
                JOIN response rb ON rb.id = h.response_b_id
                JOIN model ma ON ma.id = ra.model_id
                JOIN model mb ON mb.id = rb.model_id
                WHERE ma.id = $model_id
                OR mb.id = $model_id
            """,
                dict(model_id=model_id),
            ).df()
        return df_h2h

    @staticmethod
    def get_head_to_head_stats(project_slug: str, model_id: int) -> list[api.ModelHeadToHeadStats]:
        with ProjectService.connect(project_slug) as conn:
            df_h2h_stats = conn.execute(
                """
                WITH head_to_head_response AS (
                    SELECT
                        ra.model_id,
                        rb.model_id AS other_model_id,
                        h.judge_id,
                        CASE WHEN h.winner = 'A' THEN TRUE WHEN h.winner = 'B' THEN FALSE END AS won
                    FROM head_to_head h
                    JOIN response ra ON ra.id = h.response_a_id
                    JOIN response rb ON rb.id = h.response_b_id
                    UNION ALL
                    SELECT
                        rb.model_id,
                        ra.model_id AS other_model_id,
                        h.judge_id,
                        CASE WHEN h.winner = 'B' THEN TRUE WHEN h.winner = 'A' THEN FALSE END AS won
                    FROM head_to_head h
                    JOIN response ra ON ra.id = h.response_a_id
                    JOIN response rb ON rb.id = h.response_b_id
                )
                SELECT
                    m_other.id AS other_model_id,
                    m_other.name AS other_model_name,
                    j.id AS judge_id,
                    j.name AS judge_name,
                    SUM(IF(hr.won IS TRUE, 1, 0)) AS count_wins,
                    SUM(IF(hr.won IS FALSE, 1, 0)) AS count_losses,
                    SUM(IF(hr.won IS NULL, 1, 0)) AS count_ties
                FROM head_to_head_response hr
                JOIN judge j ON j.id = hr.judge_id
                JOIN model m ON m.id = hr.model_id
                JOIN model m_other ON m_other.id = hr.other_model_id
                WHERE m.id = $model_id
                GROUP BY m.id, m.name, m_other.id, m_other.name, j.id, j.name
            """,
                dict(model_id=model_id),
            ).df()
        return [api.ModelHeadToHeadStats(**r) for _, r in df_h2h_stats.iterrows()]



import time
from collections import defaultdict

import pandas as pd
from pydantic.dataclasses import dataclass
from loguru import logger

from autoarena.api import api
from autoarena.service.project import ProjectService


@dataclass(frozen=True)
class EloConfig:
    default_score: float = 1_000
    k: int = 4
    scale: int = 400
    base: int = 10
    n_bootstrap_rounds: int = 200


DEFAULT_ELO_CONFIG = EloConfig()


class EloService:
    @staticmethod
    def get_df_head_to_head(project_slug: str) -> pd.DataFrame:
        with ProjectService.connect(project_slug) as conn:
            return conn.execute(
                """
                SELECT
                    j.id AS judge_id,
                    j.name AS judge_name,
                    ma.id AS model_a_id,
                    ma.name AS model_a,
                    mb.id AS model_b_id,
                    mb.name AS model_b,
                    h.winner
                FROM head_to_head h
                JOIN judge j ON h.judge_id = j.id
                JOIN response ra ON h.response_a_id = ra.id
                JOIN response rb ON h.response_b_id = rb.id
                JOIN model ma ON ra.model_id = ma.id
                JOIN model mb ON rb.model_id = mb.id
                ORDER BY h.id -- ensure we are replaying head-to-heads in the order they were submitted
            """,
            ).df()

    @staticmethod
    def reseed_scores(project_slug: str, config: EloConfig = DEFAULT_ELO_CONFIG) -> None:
        df_h2h = EloService.get_df_head_to_head(project_slug)
        df_elo = EloService.compute_elo(df_h2h, config=config)  # noqa: F841
        with ProjectService.connect(project_slug) as conn:
            conn.execute(
                """
                UPDATE model
                SET elo = IFNULL(df_elo.elo, $default_elo), q025 = df_elo.q025, q975 = df_elo.q975
                FROM model m2
                LEFT JOIN df_elo ON df_elo.model = m2.name -- left join to set null values for any models without votes
                WHERE model.id = m2.id;
                """,
                dict(default_elo=config.default_score),
            )

    # most elo-related code is from https://github.com/lm-sys/FastChat/blob/main/fastchat/serve/monitor/elo_analysis.py
    @staticmethod
    def compute_elo_single(
        elo_a: float,
        elo_b: float,
        winner: api.WinnerType,
        config: EloConfig = DEFAULT_ELO_CONFIG,
    ) -> tuple[float, float]:
        expected_a = 1 / (1 + config.base ** ((elo_b - elo_a) / config.scale))
        expected_b = 1 / (1 + config.base ** ((elo_a - elo_b) / config.scale))
        score_a = 1 if winner == "A" else 0 if winner == "B" else 0.5
        elo_a += config.k * (score_a - expected_a)
        elo_b += config.k * (1 - score_a - expected_b)
        return elo_a, elo_b

    @staticmethod
    def compute_elo(df_h2h: pd.DataFrame, *, config: EloConfig = DEFAULT_ELO_CONFIG) -> pd.DataFrame:
        df_elo = EloService._compute_elo_once(df_h2h, config=config)
        df_bootstrap = EloService._get_bootstrap_result(df_h2h, config=config)
        # set Elo score to the median of the bootstrap result -- the order of the head-to-heads doesn't matter for bulk
        #  judging so the "true" score is in the middle of the differently ordered rollouts
        df_elo = df_elo.merge(df_bootstrap.quantile(0.5).rename("elo_median"), left_on="model", right_index=True)
        df_elo["elo"] = df_elo["elo_median"]
        df_elo = df_elo.merge(df_bootstrap.quantile(0.025).rename("q025"), left_on="model", right_index=True)
        df_elo = df_elo.merge(df_bootstrap.quantile(0.975).rename("q975"), left_on="model", right_index=True)
        df_elo["ci95"] = df_elo["q975"] - df_elo["q025"]
        return df_elo[["model", "elo", "q025", "q975", "ci95"]]

    @staticmethod
    def _compute_elo_once(df_h2h: pd.DataFrame, *, config: EloConfig = DEFAULT_ELO_CONFIG) -> pd.DataFrame:
        rating: dict[str, float] = defaultdict(lambda: config.default_score)
        for _, model_a, model_b, winner in df_h2h[["model_a", "model_b", "winner"]].itertuples():
            elo_a, elo_b = EloService.compute_elo_single(rating[model_a], rating[model_b], winner, config=config)
            rating[model_a] = elo_a
            rating[model_b] = elo_b
        df_elo = pd.DataFrame(dict(rating).items(), columns=["model", "elo"])
        return df_elo.sort_values(by="elo", ascending=False)

    @staticmethod
    def _get_bootstrap_result(df_h2h: pd.DataFrame, *, config: EloConfig = DEFAULT_ELO_CONFIG) -> pd.DataFrame:
        rows = []
        t_start = time.time()
        logger.info(f"Bootstrapping confidence intervals with {config.n_bootstrap_rounds} rounds...")
        for _ in range(config.n_bootstrap_rounds):
            df_h2h_tmp = df_h2h.sample(frac=1.0, replace=True)
            rows.append(EloService._compute_elo_once(df_h2h_tmp, config=config))
        logger.info(f"Bootstrapped confidence intervals in {time.time() - t_start:0.1f} seconds")
        df = pd.DataFrame([{r.model: r.elo for r in df_row.itertuples()} for df_row in rows])
        return df[df.median().sort_values(ascending=False).index]



from contextlib import asynccontextmanager
from typing import AsyncIterator

from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from loguru import logger

from autoarena.api.router import router
from autoarena.log import initialize_logger
from autoarena.service.project import ProjectService
from autoarena.store.database import get_data_directory
from autoarena.ui_router import ui_router

API_V1_STR = "/api/v1"


@asynccontextmanager
async def lifespan(_: FastAPI) -> AsyncIterator[None]:
    logger.info(f"Using data directory: '{get_data_directory()}'")
    ProjectService.migrate_all()
    logger.success("AutoArena ready")
    yield


def server() -> FastAPI:
    initialize_logger()
    app = FastAPI(title="AutoArena", lifespan=lifespan, docs_url=f"{API_V1_STR}/docs")
    app.add_middleware(
        CORSMiddleware,
        allow_origins=["*"],
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )
    app.include_router(router(), prefix=API_V1_STR)
    app.include_router(ui_router())
    return app



import os
import time

import google.generativeai as genai
from loguru import logger

from autoarena.judge.base import AutomatedJudge
from autoarena.judge.utils import get_user_prompt, JOINED_PROMPT_TEMPLATE, rate_limit


class GeminiJudge(AutomatedJudge):
    API_KEY_NAME = "GOOGLE_API_KEY"

    def __init__(self, name: str, model_name: str, system_prompt: str) -> None:
        super().__init__(name, model_name, system_prompt)
        genai.configure(api_key=os.environ.get(GeminiJudge.API_KEY_NAME, None))
        self._model = genai.GenerativeModel(model_name)

    @staticmethod
    def verify_environment() -> None:
        genai.configure(api_key=os.environ.get(GeminiJudge.API_KEY_NAME, None))
        # TODO: this takes a while, likely due to retries -- haven't figured out how to disable
        list(genai.list_models(page_size=1))

    @rate_limit(n_calls=1_000, n_seconds=60)
    def judge(self, prompt: str, response_a: str, response_b: str) -> str:
        user_prompt = get_user_prompt(prompt, response_a, response_b)
        full_prompt = JOINED_PROMPT_TEMPLATE.format(system_prompt=self.system_prompt, user_prompt=user_prompt)
        t0 = time.time()
        response = self._model.generate_content(
            full_prompt,
            generation_config=dict(max_output_tokens=self.MAX_TOKENS, temperature=0.0),
            # there's a gremlin here where some responses are still blocked for safety despite these settings
            safety_settings={
                genai.types.HarmCategory.HARM_CATEGORY_HATE_SPEECH: genai.types.HarmBlockThreshold.BLOCK_NONE,
                genai.types.HarmCategory.HARM_CATEGORY_DANGEROUS_CONTENT: genai.types.HarmBlockThreshold.BLOCK_NONE,
                genai.types.HarmCategory.HARM_CATEGORY_SEXUALLY_EXPLICIT: genai.types.HarmBlockThreshold.BLOCK_NONE,
                genai.types.HarmCategory.HARM_CATEGORY_HARASSMENT: genai.types.HarmBlockThreshold.BLOCK_NONE,
            },
        )
        self.update_usage(
            response.usage_metadata.prompt_token_count,
            response.usage_metadata.candidates_token_count,
            time.time() - t0,
        )
        if response.prompt_feedback.block_reason != 0:
            message = f"Prompt blocked by '{self.name}' safety filters: {response.prompt_feedback}. Recording as '-'"
            logger.warning(message)
            return "-"
        return response.text



from contextlib import contextmanager
from pathlib import Path

import duckdb
from loguru import logger

from autoarena.api import api
from autoarena.error import NotFoundError, MigrationError
from autoarena.store.database import get_database_connection, get_data_directory, get_available_migrations


class ProjectService:
    @staticmethod
    @contextmanager
    def connect(slug: str) -> duckdb.DuckDBPyConnection:
        path = ProjectService._slug_to_path(slug)
        if not path.exists():
            raise NotFoundError(f"File for project '{slug}' not found (expected: {path})")
        with get_database_connection(path) as conn:
            yield conn

    @staticmethod
    def get_all() -> list[api.Project]:
        paths = sorted(list(get_data_directory().glob("*.duckdb")))
        return [api.Project(slug=ProjectService._path_to_slug(p), filename=p.name, filepath=str(p)) for p in paths]

    @staticmethod
    def create_idempotent(request: api.CreateProjectRequest) -> api.Project:
        # TODO: should come up with a better way than this to have services point at one another, or remove the need
        from autoarena.service.judge import JudgeService

        data_directory = get_data_directory()
        data_directory.mkdir(parents=True, exist_ok=True)
        path = data_directory / f"{request.name}.duckdb"
        slug = ProjectService._path_to_slug(path)
        ProjectService._setup_database(path)
        JudgeService.create_human_judge(slug)
        return api.Project(slug=slug, filename=path.name, filepath=str(path))

    @staticmethod
    def delete(slug: str) -> None:
        path = ProjectService._slug_to_path(slug)
        path.unlink(missing_ok=True)
        logger.info(f"Removed file '{path}' containing project '{slug}'")

    @staticmethod
    def migrate_all() -> None:
        for project in ProjectService.get_all():
            path = Path(project.filepath)
            ProjectService._setup_database(path)
            logger.info(f"Found project '{path.relative_to(Path.cwd())}'")

    @staticmethod
    def _path_to_slug(path: Path) -> str:
        return path.stem

    @staticmethod
    def _slug_to_path(slug: str) -> Path:
        return get_data_directory() / f"{slug}.duckdb"

    @staticmethod
    def _setup_database(path: Path) -> None:
        ProjectService._migrate_to_latest(path)
        ProjectService._close_pending_tasks(path)

    @staticmethod
    def _migrate_to_latest(path: Path) -> None:
        available_migrations = get_available_migrations()
        applied_migrations = {f for (_, f) in ProjectService._get_applied_migrations(path)}
        for migration in available_migrations:
            if migration.name in applied_migrations:
                continue
            try:
                with get_database_connection(path) as conn:
                    logger.info(f"Applying migration '{migration.name}' to '{path.name}'")
                    conn.sql(migration.read_text())
                    conn.execute(
                        "INSERT INTO migration (migration_index, filename) VALUES ($index, $filename)",
                        dict(index=int(migration.name.split("__")[0]), filename=migration.name),
                    )
            except Exception as e:
                logger.error(f"Failed to apply migration '{migration.name}' to '{path.name}': {e}")
                raise MigrationError(e)

    @staticmethod
    def _get_applied_migrations(path: Path) -> list[tuple[int, str]]:
        try:
            with get_database_connection(path) as conn:
                conn.execute("SELECT migration_index, filename FROM migration ORDER BY migration_index")
                return conn.fetchall()
        except duckdb.CatalogException:
            return []  # database is new and does not have a migration table

    # TODO: restart pending tasks rather than simply terminating
    @staticmethod
    def _close_pending_tasks(path: Path) -> None:
        from autoarena.service.task import TaskService

        slug = ProjectService._path_to_slug(path)
        tasks = TaskService.get_all(slug)
        for task in tasks:
            if task.status not in {api.TaskStatus.COMPLETED, api.TaskStatus.FAILED}:
                TaskService.update(slug, task.id, "Terminated", status=api.TaskStatus.FAILED)



import time

import together

from autoarena.judge.base import AutomatedJudge
from autoarena.judge.utils import rate_limit, get_user_prompt


class TogetherJudge(AutomatedJudge):
    API_KEY_NAME = "TOGETHER_API_KEY"

    def __init__(self, name: str, model_name: str, system_prompt: str) -> None:
        super().__init__(name, model_name, system_prompt)
        self._client = together.Client()

    @staticmethod
    def verify_environment() -> None:
        together.Client().models.list()

    @rate_limit(n_calls=10, n_seconds=1, n_call_buffer=2)
    def judge(self, prompt: str, response_a: str, response_b: str) -> str:
        t0 = time.time()
        response = self._client.chat.completions.create(
            model=self.model_name,
            messages=[
                dict(role="system", content=self.system_prompt),
                dict(role="user", content=get_user_prompt(prompt, response_a, response_b)),
            ],
            max_tokens=self.MAX_TOKENS,
        )
        self.update_usage(response.usage.prompt_tokens, response.usage.completion_tokens, time.time() - t0)
        return response.choices[0].message.content



import time

import cohere

from autoarena.judge.base import AutomatedJudge
from autoarena.judge.utils import get_user_prompt, rate_limit


class CohereJudge(AutomatedJudge):
    API_KEY_NAME = "COHERE_API_KEY"  # TODO: also support "CO_API_KEY"?

    def __init__(self, name: str, model_name: str, system_prompt: str) -> None:
        super().__init__(name, model_name, system_prompt)
        self._client = cohere.Client()

    @staticmethod
    def verify_environment() -> None:
        cohere.Client().models.list()

    @rate_limit(n_calls=1_000, n_seconds=60)
    def judge(self, prompt: str, response_a: str, response_b: str) -> str:
        t0 = time.time()
        response = self._client.chat(
            model=self.model_name,
            preamble=self.system_prompt,
            message=get_user_prompt(prompt, response_a, response_b),
            max_tokens=self.MAX_TOKENS,
        )
        self.update_usage(
            int(response.meta.billed_units.input_tokens),
            int(response.meta.billed_units.output_tokens),
            time.time() - t0,
        )
        return response.text



from autoarena.api import api
from autoarena.service.task import TaskService


class FineTuningService:
    # TODO: this is just a placeholder for now
    @staticmethod
    def create_task(project_slug: str, request: api.CreateFineTuningTaskRequest) -> None:
        # 1. kick off fine-tuning job
        # 2. create judge when complete
        task_id = TaskService.create(project_slug, api.TaskType.FINE_TUNE).id
        TaskService.update(project_slug, task_id, f"Fine-tuning '{request.base_model}' base model", progress=0.1)



from fastapi import HTTPException


class BadRequestError(HTTPException):
    def __init__(self, msg: str):
        super().__init__(status_code=400, detail=msg)


class NotFoundError(HTTPException):
    def __init__(self, msg: str):
        super().__init__(status_code=404, detail=msg)


class MigrationError(RuntimeError): ...



from collections import OrderedDict
from io import BytesIO, StringIO
from typing import Optional

import pandas as pd
from fastapi import APIRouter, UploadFile, BackgroundTasks
from starlette.requests import Request
from starlette.responses import StreamingResponse

from autoarena.api import api
from autoarena.api.utils import SSEStreamingResponse
from autoarena.error import NotFoundError, BadRequestError
from autoarena.service.elo import EloService
from autoarena.service.fine_tuning import FineTuningService
from autoarena.service.head_to_head import HeadToHeadService
from autoarena.service.project import ProjectService
from autoarena.service.judge import JudgeService
from autoarena.service.task import TaskService
from autoarena.service.model import ModelService


def router() -> APIRouter:
    r = APIRouter()

    @r.get("/projects")
    def get_projects() -> list[api.Project]:
        return ProjectService.get_all()

    @r.put("/project")
    def create_project(request: api.CreateProjectRequest) -> api.Project:
        return ProjectService.create_idempotent(request)

    @r.delete("/project/{project_slug}")
    def delete_project(project_slug: str) -> None:
        return ProjectService.delete(project_slug)

    @r.get("/project/{project_slug}/models")
    def get_models(project_slug: str) -> list[api.Model]:
        return ModelService.get_all(project_slug)

    @r.get("/project/{project_slug}/models/by-judge/{judge_id}")
    def get_models_ranked_by_judge(project_slug: str, judge_id: int) -> list[api.Model]:
        return ModelService.get_all_ranked_by_judge(project_slug, judge_id)

    @r.post("/project/{project_slug}/model")
    async def upload_model_responses(
        project_slug: str,
        request: Request,
        background_tasks: BackgroundTasks,
    ) -> list[api.Model]:
        # ideally there wouldn't be this much complexity in the router, but this is a complex form to parse
        form = await request.form()
        df_response_by_model_name: dict[str, pd.DataFrame] = OrderedDict()
        for key, value in form.items():
            model_name_slug = "||model_name"
            if not key.endswith(model_name_slug):
                continue
            file: UploadFile = form[key[: -len(model_name_slug)]]
            if file.content_type != "text/csv":
                raise ValueError(f"unsupported file type: {file.content_type}")
            df_response_by_model_name[value] = pd.read_csv(BytesIO(await file.read()))
        if len(df_response_by_model_name) == 0:
            raise BadRequestError("No valid model responses in body")
        new_models = [
            ModelService.upload_responses(project_slug, model_name, df_response)
            for model_name, df_response in df_response_by_model_name.items()
        ]
        background_tasks.add_task(TaskService.auto_judge, project_slug, models=new_models)
        return new_models

    @r.get("/project/{project_slug}/model/{model_id}/responses")
    def get_model_responses(project_slug: str, model_id: int) -> list[api.ModelResponse]:
        return ModelService.get_responses(project_slug, model_id)

    # TODO: potentially remove this -- it's not intuitive to have this trigger exist at the per-model level
    @r.post("/project/{project_slug}/model/{model_id}/judge")
    def trigger_model_auto_judge(project_slug: str, model_id: int, background_tasks: BackgroundTasks) -> None:
        model = ModelService.get_by_id(project_slug, model_id)
        background_tasks.add_task(TaskService.auto_judge, project_slug, models=[model])

    @r.delete("/project/{project_slug}/model/{model_id}")
    def delete_model(project_slug: str, model_id: int, background_tasks: BackgroundTasks) -> None:
        try:
            ModelService.delete(project_slug, model_id)
            background_tasks.add_task(TaskService.recompute_leaderboard, project_slug)
        except NotFoundError:
            pass

    # async for StreamingResponses to improve speed; see https://github.com/fastapi/fastapi/issues/2302
    @r.get("/project/{project_slug}/model/{model_id}/download/responses")
    async def download_model_responses_csv(project_slug: str, model_id: int) -> StreamingResponse:
        columns = ["prompt", "response"]
        df_response = ModelService.get_df_response(project_slug, model_id)
        model_name = df_response.iloc[0].model
        stream = StringIO()
        df_response[columns].to_csv(stream, index=False)
        response = StreamingResponse(iter([stream.getvalue()]), media_type="text/csv")
        response.headers["Content-Disposition"] = f'attachment; filename="{model_name}.csv"'
        return response

    @r.get("/project/{project_slug}/model/{model_id}/download/head-to-heads")
    async def download_model_head_to_heads_csv(project_slug: str, model_id: int) -> StreamingResponse:
        columns = ["prompt", "model_a", "model_b", "response_a", "response_b", "judge", "winner"]
        df_h2h = ModelService.get_df_head_to_head(project_slug, model_id)
        model = ModelService.get_by_id(project_slug, model_id)
        stream = StringIO()
        df_h2h[columns].to_csv(stream, index=False)
        response = StreamingResponse(iter([stream.getvalue()]), media_type="text/csv")
        response.headers["Content-Disposition"] = f'attachment; filename="{model.name}-head-to-head.csv"'
        return response

    @r.get("/project/{project_slug}/model/{model_id}/head-to-head/stats")
    def get_head_to_head_stats(project_slug: str, model_id: int) -> list[api.ModelHeadToHeadStats]:
        return ModelService.get_head_to_head_stats(project_slug, model_id)

    @r.put("/project/{project_slug}/head-to-heads")
    def get_head_to_heads(project_slug: str, request: api.HeadToHeadsRequest) -> list[api.HeadToHead]:
        return HeadToHeadService.get(project_slug, request)

    @r.get("/project/{project_slug}/head-to-head/count")
    def get_head_to_head_count(project_slug: str) -> int:
        return HeadToHeadService.get_count(project_slug)

    @r.post("/project/{project_slug}/head-to-head/vote")
    def submit_head_to_head_vote(
        project_slug: str,
        request: api.HeadToHeadVoteRequest,
        background_tasks: BackgroundTasks,
    ) -> None:
        HeadToHeadService.submit_vote(project_slug, request)
        # recompute confidence intervals in the background if we aren't doing so already
        background_tasks.add_task(TaskService.recompute_leaderboard, project_slug)

    @r.get("/project/{project_slug}/tasks")
    def get_tasks(project_slug: str) -> list[api.Task]:
        return TaskService.get_all(project_slug)

    @r.get("/project/{project_slug}/task/{task_id}/stream")
    async def get_task_stream(project_slug: str, task_id: int) -> StreamingResponse:  # Iterator[api.Task]
        return SSEStreamingResponse(TaskService.get_stream(project_slug, task_id))

    @r.get("/project/{project_slug}/tasks/has-active")
    async def get_has_active_tasks_stream(
        project_slug: str,
        timeout: Optional[float] = None,
    ) -> StreamingResponse:  # Iterator[api.HasActiveTasks]
        return SSEStreamingResponse(TaskService.has_active_stream(project_slug, timeout=timeout))

    @r.delete("/project/{project_slug}/tasks/completed")
    def delete_completed(project_slug: str) -> None:
        TaskService.delete_completed(project_slug)

    @r.post("/project/{project_slug}/task/auto-judge")
    def trigger_auto_judge(
        project_slug: str,
        request: api.TriggerAutoJudgeRequest,
        background_tasks: BackgroundTasks,
    ) -> None:
        judges = [j for j in JudgeService.get_all(project_slug) if j.id in set(request.judge_ids)]
        kwargs = dict(judges=judges, fraction=request.fraction, skip_existing=request.skip_existing)
        background_tasks.add_task(TaskService.auto_judge, project_slug, **kwargs)

    @r.get("/project/{project_slug}/judges")
    def get_judges(project_slug: str) -> list[api.Judge]:
        return JudgeService.get_all(project_slug)

    @r.get("/project/{project_slug}/judge/default-system-prompt")
    def get_default_system_prompt(project_slug: str) -> str:
        return JudgeService.get_default_system_prompt()

    @r.post("/project/{project_slug}/judge")
    def create_judge(project_slug: str, request: api.CreateJudgeRequest) -> api.Judge:
        return JudgeService.create(project_slug, request)

    @r.put("/project/{project_slug}/judge/{judge_id}")
    def update_judge(project_slug: str, judge_id: int, request: api.UpdateJudgeRequest) -> api.Judge:
        return JudgeService.update(project_slug, judge_id, request)

    @r.get("/project/{project_slug}/judge/{judge_type}/can-access")
    def check_can_access(project_slug: str, judge_type: api.JudgeType) -> bool:
        return JudgeService.check_can_access(judge_type)

    @r.delete("/project/{project_slug}/judge/{judge_id}")
    def delete_judge(project_slug: str, judge_id: int, background_tasks: BackgroundTasks) -> None:
        try:
            JudgeService.delete(project_slug, judge_id)
            background_tasks.add_task(TaskService.recompute_leaderboard, project_slug)
        except NotFoundError:
            pass

    @r.put("/project/{project_slug}/elo/reseed-scores")
    def reseed_scores(project_slug: str) -> None:
        EloService.reseed_scores(project_slug)

    @r.post("/project/{project_slug}/fine-tune")
    def create_fine_tuning_task(project_slug: str, request: api.CreateFineTuningTaskRequest) -> None:
        FineTuningService.create_task(project_slug, request)

    return r



import time

import httpx
import ollama

from autoarena.judge.base import AutomatedJudge
from autoarena.judge.utils import get_user_prompt


class OllamaJudge(AutomatedJudge):
    API_KEY_NAME = None  # does not require an API key

    def __init__(self, name: str, model_name: str, system_prompt: str) -> None:
        super().__init__(name, model_name, system_prompt)
        self._client = ollama
        try:
            self._client.show(model_name)  # ensure this model exists and is pulled
        except ollama.ResponseError:
            raise RuntimeError(f"Ollama model '{model_name}' not found, try pulling it first with 'ollama pull'")
        except httpx.ConnectError:
            raise RuntimeError("Unable to connect to Ollama, ensure it is running on the same host running AutoArena")

    @staticmethod
    def verify_environment() -> None:
        try:
            ollama.list()
        except httpx.ConnectError:
            raise RuntimeError("Unable to connect to Ollama, ensure it is running on the same host running AutoArena")

    def judge(self, prompt: str, response_a: str, response_b: str) -> str:
        t0 = time.time()
        response = self._client.chat(
            model=self.model_name,
            messages=[
                dict(role="system", content=self.system_prompt),
                dict(role="user", content=get_user_prompt(prompt, response_a, response_b)),
            ],
            options=dict(temperature=0, seed=0, num_predict=self.MAX_TOKENS),
        )
        self.update_usage(response["prompt_eval_count"], response["eval_count"], time.time() - t0)
        return response["message"]["content"]



from autoarena.judge.base import AutomatedJudge

_CUSTOM_JUDGE_NAME_TO_CLASS: dict[str, type[AutomatedJudge]] = {}


def register_custom_judge_class(custom_judge_name: str, custom_judge_class: type[AutomatedJudge]) -> None:
    global _CUSTOM_JUDGE_NAME_TO_CLASS
    _CUSTOM_JUDGE_NAME_TO_CLASS[custom_judge_name] = custom_judge_class


def get_custom_judge_class(custom_judge_name: str) -> type[AutomatedJudge]:
    if custom_judge_name not in _CUSTOM_JUDGE_NAME_TO_CLASS:
        raise ValueError(f"No implementation for custom judge with name '{custom_judge_name}'")
    return _CUSTOM_JUDGE_NAME_TO_CLASS[custom_judge_name]


def clear_custom_judge_classes() -> None:
    global _CUSTOM_JUDGE_NAME_TO_CLASS
    _CUSTOM_JUDGE_NAME_TO_CLASS.clear()
